r"""Utils
=========
"""
from .finders import (
    InvalidPackageFinder,
    InvalidPathFinder,
    OptionFinder,
    RepeatedPackageFinder,
    UnsortedRequirementFinder,
)
from .tree_sitter_lsp.finders import ErrorFinder, MissingFinder

DIAGNOSTICS_FINDERS = [
    ErrorFinder(),
    MissingFinder(),
    InvalidPackageFinder(),
    InvalidPathFinder(),
    RepeatedPackageFinder(),
    UnsortedRequirementFinder(),
]
DIAGNOSTICS_FINDERS_PEP508 = [
    OptionFinder(),
    ErrorFinder(),
    MissingFinder(),
    InvalidPackageFinder(),
    InvalidPathFinder(),
    RepeatedPackageFinder(True),
    UnsortedRequirementFinder(),
]
FORMATTING_FINDER = UnsortedRequirementFinder()



r"""Format
==========
"""
from typing import Callable

from lsprotocol.types import Position, Range, TextEdit
from tree_sitter import Tree

from . import Finder


def position_2d_to_1d(source: str, position: Position) -> int:
    r"""Position 2d to 1d.

    :param source:
    :type source: str
    :param position:
    :type position: Position
    :rtype: int
    """
    return (
        sum(len(line) + 1 for line in source.splitlines()[: position.line])
        + position.character
    )


def range_2d_to_1d(source: str, region: Range) -> range:
    r"""Range 2d to 1d.

    :param source:
    :type source: str
    :param region:
    :type region: Range
    :rtype: range
    """
    return range(
        position_2d_to_1d(source, region.start),
        position_2d_to_1d(source, region.end),
    )


def apply_text_edits(text_edits: list[TextEdit], source: str) -> str:
    r"""Apply text edits.

    :param text_edits:
    :type text_edits: list[TextEdit]
    :param source:
    :type source: str
    :rtype: str
    """
    for text_edit in text_edits:
        region = range_2d_to_1d(source, text_edit.range)
        source = (
            source[: region.start] + text_edit.new_text + source[region.stop :]
        )
    return source


def format(
    paths: list[str], finder: Finder, parse: Callable[[bytes], Tree]
) -> None:
    r"""Format.

    :param paths:
    :type paths: list[str]
    :param finder:
    :type finder: Finder
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :rtype: None
    """
    for path in paths:
        with open(path, "rb") as f:
            src = f.read()
        tree = parse(src)
        finder.find_all(path, tree)
        text_edits = finder.get_text_edits()
        src = apply_text_edits(text_edits, src.decode())
        with open(path, "w") as f:
            f.write(src)



r"""Test server."""
import os

from requirements_language_server.server import RequirementsLanguageServer


class Test:
    r"""Test."""

    @staticmethod
    def test_get_dependencies_files() -> None:
        r"""Test get dependencies files.

        :rtype: None
        """
        result = RequirementsLanguageServer.get_dependencies_files(
            os.path.join(
                os.path.dirname(os.path.dirname(__file__)), "pyproject.toml"
            )
        )
        assert result != []



r"""Test package"""
from requirements_language_server import NOT_FOUND
from requirements_language_server.documents.package import search_document


class Test:
    r"""Test."""

    @staticmethod
    def test_get_document() -> None:
        r"""Test get document.

        :rtype: None
        """
        assert search_document("pip", "") != NOT_FOUND
        assert search_document("", "") == NOT_FOUND



r"""Test utils."""
import os

from tree_sitter_requirements import parse

from requirements_language_server.tree_sitter_lsp.diagnose import check
from requirements_language_server.utils import DIAGNOSTICS_FINDERS


class Test:
    r"""Test."""

    @staticmethod
    def test_check() -> None:
        r"""Test check.

        :rtype: None
        """
        result = check(
            [os.path.join(os.path.dirname(__file__), "requirements.txt.in")],
            DIAGNOSTICS_FINDERS,
            parse,
        )
        assert result > 0



r"""This module can be called by
`python -m <https://docs.python.org/3/library/__main__.html>`_.
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from contextlib import suppress
from datetime import datetime

from . import CONFIG
from . import __name__ as NAME
from . import __version__

NAME = NAME.replace("_", "-")
VERSION = rf"""{NAME} {__version__}
Copyright (C) {datetime.now().year}
Written by Wu Zhenyu
"""
EPILOG = """
Report bugs to <wuzhenyu@ustc.edu>.
"""


def get_parser() -> ArgumentParser:
    r"""Get a parser for unit test."""
    parser = ArgumentParser(
        epilog=EPILOG,
        formatter_class=RawDescriptionHelpFormatter,
    )
    with suppress(ImportError):
        import shtab

        shtab.add_argument_to(parser)
    parser.add_argument("--version", version=VERSION, action="version")
    parser.add_argument(
        "--print-config",
        choices=["template", "cache", "all"],
        help="print config value",
    )
    parser.add_argument(
        "--generate-cache",
        action="store_true",
        help="generate cache",
    )
    parser.add_argument(
        "--check",
        nargs="*",
        default=[],
        help="check file's errors and warnings",
    )
    parser.add_argument(
        "--format",
        nargs="*",
        default=[],
        help="format files",
    )
    parser.add_argument(
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="when to display color",
    )
    return parser


def main() -> None:
    r"""Parse arguments and provide shell completions."""
    parser = get_parser()
    args = parser.parse_args()

    from tree_sitter_requirements import parse

    from .tree_sitter_lsp.diagnose import check
    from .tree_sitter_lsp.format import format
    from .utils import DIAGNOSTICS_FINDERS, FORMATTING_FINDER

    format(args.format, FORMATTING_FINDER, parse)
    result = check(args.check, DIAGNOSTICS_FINDERS, parse, args.color)
    if args.format or args.check:
        exit(result)
    if args.print_config:
        print(
            CONFIG.get(
                args.print_config,
                "\n".join([k + ": " + v for k, v in CONFIG.items()]),
            )
        )
        return None
    if args.generate_cache:
        from .documents.package import generate_cache

        generate_cache()
        return None

    from .server import RequirementsLanguageServer

    RequirementsLanguageServer(NAME, __version__).start_io()


if __name__ == "__main__":
    main()



r"""Tree-sitter LSP
===================
"""
import os
from copy import deepcopy
from typing import Any

from jinja2 import Template
from lsprotocol.types import (
    Diagnostic,
    DiagnosticSeverity,
    DocumentLink,
    Location,
    Position,
    Range,
    TextEdit,
)
from pygls.uris import to_fs_path
from tree_sitter import Node, Tree, TreeCursor

# maximum of recursive search
LEVEL = 5


class UNI:
    r"""Unified node identifier."""

    def __init__(self, uri: str, node: Node) -> None:
        r"""Init.

        :param uri:
        :type uri: str
        :param node:
        :type node: Node
        :rtype: None
        """
        self.uri = uri
        self.node = node

    def __repr__(self) -> str:
        r"""Repr.

        :rtype: str
        """
        return f"{self.get_text()}@{self.uri}:{self.node.start_point[0] + 1}:{self.node.start_point[1] + 1}-{self.node.end_point[0] + 1}:{self.node.end_point[1]}"

    def __eq__(self, that: "UNI") -> bool:
        r"""Eq.

        :param that:
        :type that: UNI
        :rtype: bool
        """
        return self.node == that.node

    def get_text(self) -> str:
        r"""Get text.

        :rtype: str
        """
        return self.node2text(self.node)

    @staticmethod
    def node2text(node: Node) -> str:
        r"""Node2text.

        :param node:
        :type node: Node
        :rtype: str
        """
        return node.text.decode()

    def get_location(self) -> Location:
        r"""Get location.

        :rtype: Location
        """
        return Location(self.uri, self.get_range())

    def get_range(self) -> Range:
        r"""Get range.

        :rtype: Range
        """
        return self.node2range(self.node)

    @staticmethod
    def node2range(node: Node) -> Range:
        r"""Node2range.

        :param node:
        :type node: Node
        :rtype: Range
        """
        return Range(Position(*node.start_point), Position(*node.end_point))

    def get_path(self) -> str:
        r"""Get path.

        :rtype: str
        """
        return self.uri2path(self.uri)

    @staticmethod
    def uri2path(uri: str) -> str:
        r"""Uri2path.

        :param uri:
        :type uri: str
        :rtype: str
        """
        if path := to_fs_path(uri):
            return path
        raise TypeError

    def get_diagnostic(
        self,
        message: str,
        severity: DiagnosticSeverity,
        **kwargs: Any,
    ) -> Diagnostic:
        r"""Get diagnostic.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :param kwargs:
        :type kwargs: Any
        :rtype: Diagnostic
        """
        _range = self.get_range()
        _range.end.character -= 1
        return Diagnostic(
            _range,
            Template(message).render(uni=self, **kwargs),
            severity,
        )

    def get_text_edit(self, new_text: str) -> TextEdit:
        r"""Get text edit.

        :param new_text:
        :type new_text: str
        :rtype: TextEdit
        """
        return TextEdit(self.get_range(), new_text)

    def get_document_link(self, target: str, **kwargs) -> DocumentLink:
        r"""Get document link.

        :param target:
        :type target: str
        :param kwargs:
        :rtype: DocumentLink
        """
        return DocumentLink(
            self.get_range(),
            Template(target).render(uni=self, **kwargs),
        )

    @staticmethod
    def join(path, text) -> str:
        r"""Join.

        :param path:
        :param text:
        :rtype: str
        """
        return os.path.join(os.path.dirname(path), text)


class Finder:
    r"""Finder."""

    def __init__(
        self,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        self.level = 0
        self.message = message
        self.severity = severity
        self.reset()

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return True

    def __and__(self, second: "Finder") -> "Finder":
        r"""And.

        :param second:
        :type second: Finder
        :rtype: "Finder"
        """
        finder = deepcopy(self)
        finder.__call__ = lambda uni: self(uni) and second(uni)
        return finder

    def __or__(self, second: "Finder") -> "Finder":
        r"""Or.

        :param second:
        :type second: Finder
        :rtype: "Finder"
        """
        finder = deepcopy(self)
        finder.__call__ = lambda uni: self(uni) or second(uni)
        return finder

    def __minus__(self, second: "Finder") -> "Finder":
        r"""Minus.

        :param second:
        :type second: Finder
        :rtype: "Finder"
        """
        finder = deepcopy(self)
        finder.__call__ = lambda uni: self(uni) and not second(uni)
        return finder

    def is_include_node(self, node: Node) -> bool:
        r"""Is include node.

        :param node:
        :type node: Node
        :rtype: bool
        """
        return False

    def parse(self, code: bytes) -> Tree:
        r"""Parse.

        :param code:
        :type code: bytes
        :rtype: Tree
        """
        raise NotImplementedError

    def uri2tree(self, uri: str) -> Tree | None:
        r"""Convert URI to tree.

        :param uri:
        :type uri: str
        :rtype: Tree | None
        """
        path = UNI.uri2path(uri)
        if not os.path.exists(path):
            return None
        with open(path, "rb") as f:
            code = f.read()
        return self.parse(code)

    def uni2uri(self, uni: UNI) -> str:
        r"""Convert UNI to URI.

        :param uni:
        :type uni: UNI
        :rtype: str
        """
        return uni.join(uni.uri, uni.get_text())

    def uni2path(self, uni: UNI) -> str:
        r"""Convert UNI to path.

        :param self:
        :param uni:
        :type uni: UNI
        :rtype: str
        """
        uri = self.uni2uri(uni)
        return UNI.uri2path(uri)

    def move_cursor(
        self, uri: str, cursor: TreeCursor, is_all: bool = False
    ) -> str:
        r"""Move cursor.

        :param uri:
        :type uri: str
        :param cursor:
        :type cursor: TreeCursor
        :param is_all:
        :type is_all: bool
        :rtype: str
        """
        while self(UNI(uri, cursor.node)) is False:
            if self.is_include_node(cursor.node) and self.level < LEVEL:
                self.level += 1
                old_uri = uri
                uri = self.uni2uri(UNI(uri, cursor.node))
                tree = self.uri2tree(uri)
                if tree is not None:
                    if is_all:
                        self.find_all(uri, tree)
                    else:
                        self.find(uri, tree)
                uri = old_uri
                self.level -= 1
            if cursor.node.child_count > 0:
                cursor.goto_first_child()
                continue
            while cursor.node.next_sibling is None:
                cursor.goto_parent()
                # when cannot find new nodes, return
                if cursor.node.parent is None:
                    return ""
            cursor.goto_next_sibling()
        return uri

    def reset(self) -> None:
        r"""Reset.

        :rtype: None
        """
        pass

    def find(self, uri: str, tree: Tree | None = None) -> UNI | None:
        r"""Find.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :rtype: UNI | None
        """
        self.reset()
        if tree is None:
            tree = self.uri2tree(uri)
        if tree is None:
            return None
        cursor = tree.walk()
        if uri := self.move_cursor(uri, cursor, False):
            return UNI(uri, cursor.node)
        else:
            return None

    def find_all(self, uri: str, tree: Tree | None = None) -> list[UNI]:
        r"""Find all.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :rtype: list[UNI]
        """
        self.reset()
        if tree is None:
            tree = self.uri2tree(uri)
        if tree is None:
            return []
        cursor = tree.walk()
        unis = []
        while True:
            if uri := self.move_cursor(uri, cursor, True):
                unis += [UNI(uri, cursor.node)]
            while cursor.node.next_sibling is None:
                cursor.goto_parent()
                # when cannot find new nodes, return
                if cursor.node.parent is None:
                    return unis
            cursor.goto_next_sibling()

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        return uni.get_diagnostic(self.message, self.severity)

    def unis2diagnostics(self, unis: list[UNI]) -> list[Diagnostic]:
        r"""Unis2diagnostics.

        :param unis:
        :type unis: list[UNI]
        :rtype: list[Diagnostic]
        """
        return [self.uni2diagnostic(uni) for uni in unis]

    def get_text_edits(self) -> list[TextEdit]:
        r"""Get text edits.

        :param self:
        :rtype: list[TextEdit]
        """
        raise NotImplementedError



r"""Documents
=============
"""
import json
import os

from .. import CACHE


def get_documents() -> dict[str, str]:
    r"""Get documents.

    :rtype: dict[str, str]
    """
    documents = {}
    if os.path.exists(CACHE):
        with open(CACHE, "r") as f:
            documents: dict[str, str] = json.load(f)
    return documents



r"""Option
==========
"""
from pip._internal.commands import create_command

# https://pip.pypa.io/en/stable/reference/requirements-file-format/#supported-options
WHITELIST = [
    "-i",
    "--index-url",
    "--extra-index-url",
    "--no-index",
    "-c",
    "--constraint",
    "-r",
    "--requirement",
    "-e",
    "--editable",
    "-f",
    "--find-links",
    "--no-binary",
    "--only-binary",
    "--prefer-binary",
    "--require-hashes",
    "--pre",
    "--trusted-host",
    "--use-feature",
    "--global-option",
    "--config-settings",
    "--hash",
]

OPTIONS_WITH_EQUAL = {
    opt
    + ("=" if option.nargs and opt in option._long_opts else ""): option.help
    if option.help
    else ""
    for option in create_command("install").parser.option_list_all
    if (option._short_opts + option._long_opts)[0] in WHITELIST
    for opt in option._short_opts + option._long_opts
}

OPTIONS = {opt.rstrip("="): doc for opt, doc in OPTIONS_WITH_EQUAL.items()}



r"""Package
===========
"""
import json

from jinja2 import Template
from pip._internal.commands.show import search_packages_info
from pip._internal.metadata import get_environment

from .. import CACHE, NOT_FOUND, TEMPLATE


def search_document(pkgname: str, template: str) -> str:
    r"""Search document.

    :param pkgname:
    :type pkgname: str
    :param template:
    :type template: str
    :rtype: str
    """
    try:
        info = list(search_packages_info([pkgname]))[0]
        output = Template(template).render(info=info)
    except IndexError:
        output = NOT_FOUND
    return output


def generate_cache() -> None:
    r"""Generate cache.

    :rtype: None
    """
    documents = {
        pkgname: search_document(pkgname, TEMPLATE)
        for pkgname in [
            x.canonical_name
            for x in get_environment(None).iter_installed_distributions()
        ]
    }
    with open(CACHE, "w") as f:
        json.dump(documents, f)



r"""Diagnose
============
"""
import sys
from typing import Callable, Literal

from lsprotocol.types import Diagnostic, DiagnosticSeverity
from tree_sitter import Tree

from . import Finder


def get_diagnostics(
    finders: list[Finder], uri: str, tree: Tree
) -> list[Diagnostic]:
    r"""Get diagnostics.

    :param finders:
    :type finders: list[Finder]
    :param uri:
    :type uri: str
    :param tree:
    :type tree: Tree
    :rtype: list[Diagnostic]
    """
    return [
        diagnostic
        for finder in finders
        for diagnostic in finder.unis2diagnostics(finder.find_all(uri, tree))
    ]


def count_level(
    diagnostics: list[Diagnostic],
    level: DiagnosticSeverity = DiagnosticSeverity.Warning,
) -> int:
    r"""Count level.

    :param diagnostics:
    :type diagnostics: list[Diagnostic]
    :param level:
    :type level: DiagnosticSeverity
    :rtype: int
    """
    return len(
        [
            diagnostic
            for diagnostic in diagnostics
            if diagnostic.severity and diagnostic.severity <= level
        ]
    )


class _Colorama:
    """Colorama."""

    def __getattribute__(self, _: str) -> str:
        """Getattribute.

        :param _:
        :type _: str
        :rtype: str
        """
        return ""


def diagnostics2linter_messages(
    path: str,
    diagnostics: list[Diagnostic],
    color: Literal["auto", "always", "never"] = "auto",
    colors: list[str] | None = None,
) -> list[str]:
    r"""Diagnostics2linter messages.

    :param path:
    :type path: str
    :param diagnostics:
    :type diagnostics: list[Diagnostic]
    :param color:
    :type color: Literal["auto", "always", "never"]
    :param colors:
    :type colors: list[str] | None
    :rtype: list[str]
    """
    from colorama import Fore, init

    init()
    if not sys.stdout.isatty() and color == "auto" or color == "never":
        Fore = _Colorama()
    if colors is None:
        colors = [Fore.RESET, Fore.RED, Fore.YELLOW, Fore.BLUE, Fore.GREEN]
    return [
        f"{Fore.MAGENTA}{path}{Fore.RESET}:{Fore.CYAN}{diagnostic.range.start.line + 1}:{diagnostic.range.start.character + 1}{Fore.RESET}-{Fore.CYAN}{diagnostic.range.end.line + 1}:{diagnostic.range.end.character + 1}{Fore.RESET}:{colors[diagnostic.severity if diagnostic.severity else 0]}{str(diagnostic.severity).split('.')[-1].lower()}{Fore.RESET}: {diagnostic.message}"
        for diagnostic in diagnostics
    ]


def check(
    paths: list[str],
    finders: list[Finder],
    parse: Callable[[bytes], Tree],
    color: Literal["auto", "always", "never"] = "auto",
) -> int:
    r"""Check.

    :param paths:
    :type paths: list[str]
    :param finders:
    :type finders: list[Finder]
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :param color:
    :type color: Literal["auto", "always", "never"]
    :rtype: int
    """
    count = 0
    lines = []
    for path in paths:
        with open(path, "rb") as f:
            src = f.read()
        tree = parse(src)
        diagnostics = get_diagnostics(finders, path, tree)
        count += count_level(diagnostics)
        lines += diagnostics2linter_messages(path, diagnostics, color)
    print("\n".join(lines))
    return count



r"""Complete
============
"""
import os
from glob import glob
from pathlib import Path

from lsprotocol.types import CompletionItem, CompletionItemKind, CompletionList

from . import UNI


def get_completion_list_by_uri(
    uri: str, text: str = "", expr: str = "*"
) -> CompletionList:
    r"""Get completion list by uri.

    :param uri:
    :type uri: str
    :param text:
    :type text: str
    :param expr:
    :type expr: str
    :rtype: CompletionList
    """
    dirname = os.path.dirname(UNI.uri2path(uri))
    return CompletionList(
        False,
        [
            CompletionItem(
                x.rpartition(dirname + os.path.sep)[-1],
                kind=CompletionItemKind.File
                if os.path.isfile(x)
                else CompletionItemKind.Folder,
                documentation=Path(x).read_text()
                if os.path.isfile(x)
                else "\n".join(os.listdir(x)),
                insert_text=x.rpartition(dirname + os.path.sep)[-1],
            )
            for x in [
                file + ("" if os.path.isfile(file) else os.path.sep)
                for file in glob(
                    os.path.join(dirname, text + f"**{os.path.sep}" + expr),
                    recursive=True,
                )
            ]
        ],
    )



r"""Finders
===========
"""
import os

from lsprotocol.types import (
    Diagnostic,
    DiagnosticSeverity,
    Location,
    Position,
    Range,
    TextEdit,
)
from tree_sitter import Node

from . import UNI, Finder


class MissingFinder(Finder):
    r"""Missingfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: missing",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.child_count == 0 and node.is_missing


class ErrorFinder(Finder):
    r"""Errorfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: error",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.child_count == 0 and node.has_error


class NotFileFinder(Finder):
    r"""NotFilefinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such file or directory",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        path = self.uni2path(uni)
        return not (os.path.isfile(path) or os.path.isdir(path))


class RepeatedFinder(Finder):
    r"""Repeatedfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: is repeated on {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def reset(self) -> None:
        r"""Reset.

        :rtype: None
        """
        self.unis = []
        self.uni_pairs = []

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return True

    def compare(self, uni: UNI, _uni: UNI) -> bool:
        r"""Compare.

        :param uni:
        :type uni: UNI
        :param _uni:
        :type _uni: UNI
        :rtype: bool
        """
        return uni.node.text == _uni.node.text

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        if self.filter(uni) is False:
            return False
        for _uni in self.unis:
            if self.compare(uni, _uni):
                self.uni_pairs += [[uni, _uni]]
                return True
        self.unis += [uni]
        return False

    def get_definitions(self, uni: UNI) -> list[Location]:
        r"""Get definitions.

        :param uni:
        :type uni: UNI
        :rtype: list[Location]
        """
        for uni_, _uni in self.uni_pairs:
            # cache hit
            if uni == uni_:
                return [_uni.get_location()]
        return []

    def get_references(self, uni: UNI) -> list[Location]:
        r"""Get references.

        :param uni:
        :type uni: UNI
        :rtype: list[Location]
        """
        locations = []
        for uni_, _uni in self.uni_pairs:
            # cache hit
            if uni == _uni:
                locations += [uni_.get_location()]
        return locations

    def get_text_edits(self) -> list[TextEdit]:
        r"""Get text edits.

        :rtype: list[TextEdit]
        """
        edits = []
        for uni, _uni in self.uni_pairs:
            # swap 2 unis
            edits += [
                uni.get_text_edit(_uni.get_text()),
                _uni.get_text_edit(uni.get_text()),
            ]
        return edits

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        for uni_, _uni in self.uni_pairs:
            if uni == uni_:
                return uni.get_diagnostic(
                    self.message, self.severity, _uni=_uni
                )
        return uni.get_diagnostic(self.message, self.severity)


class UnsortedFinder(RepeatedFinder):
    r"""Unsortedfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: is unsorted due to {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def compare(self, uni: UNI, _uni: UNI) -> bool:
        r"""Compare.

        :param uni:
        :type uni: UNI
        :param _uni:
        :type _uni: UNI
        :rtype: bool
        """
        return uni.node.text < _uni.node.text


class TypeFinder(Finder):
    r"""Typefinder."""

    def __init__(
        self,
        _type: str,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ) -> None:
        r"""Init.

        :param self:
        :param _type:
        :type _type: str
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.type = _type

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.type == self.type


class PositionFinder(Finder):
    r"""Positionfinder."""

    def __init__(self, position: Position) -> None:
        r"""Init.

        :param position:
        :type position: Position
        :rtype: None
        """
        super().__init__()
        self.position = position

    @staticmethod
    def belong(position: Position, node: Node) -> bool:
        r"""Belong.

        :param position:
        :type position: Position
        :param node:
        :type node: Node
        :rtype: bool
        """
        return (
            Position(*node.start_point)
            <= position
            <= Position(*node.end_point)
        )

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.child_count == 0 and self.belong(self.position, node)


class RangeFinder(Finder):
    r"""Rangefinder."""

    def __init__(self, _range: Range) -> None:
        r"""Init.

        :param _range:
        :type _range: Range
        :rtype: None
        """
        super().__init__()
        self.range = _range

    @staticmethod
    def equal(_range: Range, node: Node) -> bool:
        r"""Equal.

        :param _range:
        :type _range: Range
        :param node:
        :type node: Node
        :rtype: bool
        """
        return _range.start == Position(
            *node.start_point
        ) and _range.end == Position(*node.end_point)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return self.equal(self.range, node)



r"""Provide ``__version__`` for
`importlib.metadata.version() <https://docs.python.org/3/library/importlib.metadata.html#distribution-versions>`_.
"""
import os

from platformdirs import user_cache_dir, user_config_dir

try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:  # for setuptools-generate
    __version__ = "rolling"
    __version_tuple__ = (0, 0, 0, __version__, "")

__all__ = ["__version__", "__version_tuple__"]

NOT_FOUND = "Not found installed package!"
PATH = os.path.join(user_config_dir("pip"), "template.md.j2")
CACHE = os.path.join(user_cache_dir("pip"), "pip.json")
CONFIG = {
    "template": PATH,
    "cache": CACHE,
}
if not os.path.exists(PATH):
    PATH = os.path.join(
        os.path.join(
            os.path.join(os.path.dirname(__file__), "assets"), "jinja2"
        ),
        "template.md.j2",
    )
with open(PATH, "r") as f:
    TEMPLATE = f.read()



# type: ignore
r"""Server
==========
"""
import os
from typing import Any

import tree_sitter_requirements as requirements
from lsprotocol.types import (
    INITIALIZE,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DOCUMENT_LINK,
    TEXT_DOCUMENT_FORMATTING,
    TEXT_DOCUMENT_HOVER,
    TEXT_DOCUMENT_REFERENCES,
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionParams,
    DidChangeTextDocumentParams,
    DocumentFormattingParams,
    DocumentLink,
    DocumentLinkParams,
    Hover,
    InitializeParams,
    Location,
    MarkupContent,
    MarkupKind,
    TextDocumentPositionParams,
    TextEdit,
)
from pip_cache import get_package_names
from pygls.server import LanguageServer

from . import NOT_FOUND, TEMPLATE
from .documents import get_documents
from .documents.option import OPTIONS, OPTIONS_WITH_EQUAL
from .finders import InvalidPackageFinder, RepeatedPackageFinder
from .tree_sitter_lsp import UNI
from .tree_sitter_lsp.complete import get_completion_list_by_uri
from .tree_sitter_lsp.diagnose import get_diagnostics
from .tree_sitter_lsp.finders import PositionFinder, TypeFinder
from .utils import (
    DIAGNOSTICS_FINDERS,
    DIAGNOSTICS_FINDERS_PEP508,
    FORMATTING_FINDER,
)

try:
    import tomllib as tomli
except ImportError:
    import tomli


class RequirementsLanguageServer(LanguageServer):
    r"""Requirements language server."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        r"""Init.

        :param self:
        :param args:
        :type args: Any
        :param kwargs:
        :type kwargs: Any
        :rtype: None
        """
        super().__init__(*args, **kwargs)
        self.template = ""
        self.documents = {}
        self.trees = {}

        @self.feature(INITIALIZE)
        def initialize(params: InitializeParams) -> None:
            r"""Initialize.

            :param params:
            :type params: InitializeParams
            :rtype: None
            """
            opts = params.initialization_options
            self.template = getattr(opts, "template", TEMPLATE)
            self.documents = get_documents()

        @self.feature(TEXT_DOCUMENT_DID_OPEN)
        @self.feature(TEXT_DOCUMENT_DID_CHANGE)
        def did_change(params: DidChangeTextDocumentParams) -> None:
            r"""Did change.

            :param params:
            :type params: DidChangeTextDocumentParams
            :rtype: None
            """
            document = self.workspace.get_document(params.text_document.uri)
            self.trees[document.uri] = requirements.parse(document.source)
            if self.is_pep508(document.uri):
                finders = DIAGNOSTICS_FINDERS_PEP508
            else:
                finders = DIAGNOSTICS_FINDERS
            diagnostics = get_diagnostics(
                finders,
                document.uri,
                self.trees[document.uri],
            )
            self.publish_diagnostics(params.text_document.uri, diagnostics)

        @self.feature(TEXT_DOCUMENT_FORMATTING)
        def format(params: DocumentFormattingParams) -> list[TextEdit]:
            r"""Format.

            :param params:
            :type params: DocumentFormattingParams
            :rtype: list[TextEdit]
            """
            document = self.workspace.get_document(params.text_document.uri)
            finder = FORMATTING_FINDER
            finder.find_all(document.uri, self.trees[document.uri])
            return finder.get_text_edits()

        @self.feature(TEXT_DOCUMENT_DEFINITION)
        def definition(params: TextDocumentPositionParams) -> list[Location]:
            r"""Get definition.

            :param params:
            :type params: TextDocumentPositionParams
            :rtype: list[Location]
            """
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return []
            finder = RepeatedPackageFinder()
            finder.find_all(document.uri, self.trees[document.uri])
            return finder.get_definitions(uni)

        @self.feature(TEXT_DOCUMENT_REFERENCES)
        def references(params: TextDocumentPositionParams) -> list[Location]:
            r"""Get references.

            :param params:
            :type params: TextDocumentPositionParams
            :rtype: list[Location]
            """
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return []
            finder = RepeatedPackageFinder()
            finder.find_all(document.uri, self.trees[document.uri])
            return finder.get_references(uni)

        @self.feature(TEXT_DOCUMENT_DOCUMENT_LINK)
        def document_link(params: DocumentLinkParams) -> list[DocumentLink]:
            r"""Get document links.

            :param params:
            :type params: DocumentLinkParams
            :rtype: list[DocumentLink]
            """
            document = self.workspace.get_document(params.text_document.uri)
            finder = InvalidPackageFinder()
            return [
                uni.get_document_link(
                    "https://pypi.org/project/{{uni.get_text()}}"
                )
                for uni in TypeFinder("package").find_all(
                    document.uri, self.trees[document.uri]
                )
                if finder(uni) is False
            ]

        @self.feature(TEXT_DOCUMENT_HOVER)
        def hover(params: TextDocumentPositionParams) -> Hover | None:
            r"""Hover.

            :param params:
            :type params: TextDocumentPositionParams
            :rtype: Hover | None
            """
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return None
            text = uni.get_text()
            if uni.node.type == "option":
                return Hover(
                    MarkupContent(MarkupKind.PlainText, OPTIONS.get(text, "")),
                    uni.get_range(),
                )
            if uni.node.type == "package":
                doc = self.documents.get(text, NOT_FOUND)
                # if cache is outdated
                if doc == NOT_FOUND:
                    from .documents.package import search_document

                    doc = search_document(text, TEMPLATE)
                if not doc:
                    return None
                return Hover(
                    MarkupContent(MarkupKind.Markdown, doc),
                    uni.get_range(),
                )

        @self.feature(TEXT_DOCUMENT_COMPLETION)
        def completions(params: CompletionParams) -> CompletionList:
            r"""Completions.

            :param params:
            :type params: CompletionParams
            :rtype: CompletionList
            """
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return CompletionList(False, [])
            text = uni.get_text()

            if uni.node.type == "package":
                return CompletionList(
                    False,
                    [
                        CompletionItem(
                            x,
                            kind=CompletionItemKind.Module,
                            # even if cache is outdated,
                            # we still don't find because it is too slow
                            documentation=MarkupContent(
                                kind=MarkupKind.Markdown,
                                value=self.documents.get(x, NOT_FOUND),
                            ),
                            insert_text=x,
                        )
                        for x in get_package_names(text)
                    ],
                )
            # uni.node.type != "option" due to incomplete
            if text.startswith("-"):
                return CompletionList(
                    False,
                    [
                        CompletionItem(
                            x,
                            kind=CompletionItemKind.Keyword,
                            documentation=OPTIONS_WITH_EQUAL[x],
                            insert_text=x,
                        )
                        for x in OPTIONS_WITH_EQUAL
                        if x.startswith(text)
                    ],
                )
            document = self.workspace.get_document(params.text_document.uri)
            return get_completion_list_by_uri(document.uri, text, "*.txt")

    def is_pep508(self, uri: str) -> bool:
        r"""Is pep508.

        :param uri:
        :type uri: str
        :rtype: bool
        """
        pyproject_uri = os.path.join(self.workspace.root_uri, "pyproject.toml")
        document = self.workspace.get_document(pyproject_uri)
        pyproject_path = UNI.uri2path(document.uri)
        path = UNI.uri2path(uri)
        if path in self.get_dependencies_files(pyproject_path):
            return True
        return False

    @staticmethod
    def get_dependencies_files(pyproject_path: str) -> list[str]:
        r"""Get dependencies files.

        :param pyproject_path:
        :type pyproject_path: str
        :rtype: list[str]
        """
        if not os.access(pyproject_path, os.R_OK):
            return []
        with open(pyproject_path, "rb") as f:
            data = tomli.load(f)
        dynamic = data.get("tool", {}).get("setuptools", {}).get("dynamic", {})
        files = []
        file = dynamic.get("dependencies", {}).get("file")
        if file:
            files += [
                os.path.abspath(
                    os.path.realpath(UNI.join(pyproject_path, file))
                )
            ]
        for value in dynamic.get("optional-dependencies", {}).values():
            file = value.get("file")
            if file:
                files += [
                    os.path.abspath(
                        os.path.realpath(UNI.join(pyproject_path, file))
                    )
                ]
        return files



r"""Finders
===========
"""
import os
from typing import Literal

import tree_sitter_requirements as requirements
from lsprotocol.types import Diagnostic, DiagnosticSeverity
from pip_cache import get_package_names
from tree_sitter import Node, Tree

from .tree_sitter_lsp import UNI, Finder
from .tree_sitter_lsp.finders import RepeatedFinder, TypeFinder, UnsortedFinder


class InvalidPackageFinder(Finder):
    r"""Invalidpackagefinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such package",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        text = uni.get_text()
        return uni.node.type == "package" and text not in get_package_names(
            text
        )


class InvalidPathFinder(Finder):
    r"""Invalidpathfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such {{_type}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    @staticmethod
    def get_option(uni: UNI) -> str:
        r"""Get option.

        :param uni:
        :type uni: UNI
        :rtype: str
        """
        option = ""
        if children := getattr(uni.node.parent, "children", None):
            if len(children) > 0:
                option = UNI.node2text(children[0])
        return option

    @staticmethod
    def get_option_type(option: str) -> Literal["file", "dir", ""]:
        r"""Get option type.

        :param option:
        :type option: str
        :rtype: Literal["file", "dir", ""]
        """
        if option in ["-r", "--requirement", "-c", "--constraint"]:
            return "file"
        if option in ["-e", "--editable"]:
            return "dir"
        return ""

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        path = self.uni2path(uni)
        option = self.get_option(uni)
        return uni.node.type == "path" and not (
            os.path.isfile(path)
            and self.get_option_type(option) == "file"
            or os.path.isdir(path)
            and self.get_option_type(option) == "dir"
        )

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        return uni.get_diagnostic(
            self.message,
            self.severity,
            _type=self.get_option_type(self.get_option(uni)),
        )


class OptionFinder(TypeFinder):
    r"""Optionfinder."""

    def __init__(
        self,
        message: str = "setuptools only supports PEP508",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__("option", message, severity)


class RepeatedPackageFinder(RepeatedFinder):
    r"""Repeatedpackagefinder."""

    def __init__(
        self,
        pep508: bool = False,
        message: str = "{{uni.get_text()}}: is repeated on {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param self:
        :param pep508:
        :type pep508: bool
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.pep508 = pep508

    def is_include_node(self, node: Node) -> bool:
        r"""Is include node.

        :param node:
        :type node: Node
        :rtype: bool
        """
        return node.type == "path" and not self.pep508

    def parse(self, code: bytes) -> Tree:
        r"""Parse.

        :param code:
        :type code: bytes
        :rtype: Tree
        """
        return requirements.parse(code)

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return uni.node.type == "package"


class UnsortedRequirementFinder(UnsortedFinder):
    r"""Unsortedrequirementfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: is unsorted due to {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return uni.node.type == "requirement"
